import torch

def get_precision_by_string(key):
    '''
    Get the precision for data and model.
    '''
    if key in ['float32', 'float', 'single']:
        return torch.float32
    elif key in ['float64', 'double']:
        return torch.float64
    elif key in ['float16', 'half']:
        return torch.float16
    else:
        raise ValueError(f'precision {key} is not supported')


from torch.optim import Adam, SGD, RMSprop, AdamW
from torch.optim.lr_scheduler import ReduceLROnPlateau, LambdaLR, OneCycleLR, LinearLR, CosineAnnealingLR, ChainedScheduler


def get_optimizer_by_string(optimizer_name, parameters, **kwargs):
    """Get optimizer by string.

    Parameters
    ----------
    optimizer_name: str
        Name of the optimizer.

    parameters: iterable
        Parameters to be optimized.

    kwargs: dict
        Keyword arguments.

    Returns
    -------
    torch.optim.Optimizer
        Optimizer.

    """
    if optimizer_name == 'adam':
        optimizer = Adam(parameters, **kwargs)
    elif optimizer_name == 'sgd':
        optimizer = SGD(parameters, **kwargs)
    elif optimizer_name == 'rmsprop':
        optimizer = RMSprop(parameters, **kwargs)
    elif optimizer_name == 'adamw':
        optimizer = AdamW(parameters, **kwargs)
    else:
        raise ValueError(f'optimizer {optimizer_name} is not supported')
    return optimizer


def get_scheduler_by_string(scheduler_list, optimizer):
    """Get scheduler by string.

    Parameters
    ----------
    scheduler_list: list[(str, dict)]
        List of scheduler names and keyword arguments.

    optimizer: torch.optim.Optimizer
        Optimizer.

    Returns
    -------
    torch.optim.lr_scheduler._LRScheduler
        Scheduler.

    """
    if scheduler_list is None:
        return None
    scheduler = []
    for scheduler_name, scheduler_kwargs in scheduler_list:
        if scheduler_name == 'plateau':
            scheduler.append(ReduceLROnPlateau(optimizer, **scheduler_kwargs))
        elif scheduler_name == 'lambda':
            scheduler.append(LambdaLR(optimizer, **scheduler_kwargs))
        elif scheduler_name == 'onecycle':
            scheduler.append(OneCycleLR(optimizer, **scheduler_kwargs))
        elif scheduler_name == 'linear':
            scheduler.append(LinearLR(optimizer, **scheduler_kwargs))
        elif scheduler_name == 'cosine':
            scheduler.append(CosineAnnealingLR(optimizer, **scheduler_kwargs))
        else:
            raise ValueError(f'scheduler {scheduler_name} is not supported')
    if len(scheduler) == 1:
        return scheduler[0]
    else:
        return ChainedScheduler(scheduler)


import torch
from torch import nn
from torch.autograd import grad
from torch_geometric.utils import scatter

from newtonnet.layers.scalers import ScaleShift



def get_output_by_string(key, n_features=None, activation=None):
    if key == 'energy':
        output_layer = EnergyOutput(n_features, activation)
    elif key == 'gradient_force':
        output_layer = GradientForceOutput()
    elif key == 'direct_force':
        output_layer = DirectForceOutput(n_features, activation)
    elif key == 'hessian':
        output_layer = HessianOutput()
    elif key == 'virial':
        output_layer = VirialOutput()
    else:
        raise NotImplementedError(f'Output type {key} is not implemented yet')
    return output_layer

def get_aggregator_by_string(key):
    if key == 'energy':
        aggregator = SumAggregator()
    elif key == 'gradient_force':
        aggregator = NullAggregator()
    elif key == 'direct_force':
        aggregator = NullAggregator()
    elif key == 'hessian':
        aggregator = NullAggregator()
    elif key == 'virial':
        aggregator = NullAggregator()
    else:
        raise NotImplementedError(f'Aggregate type {key} is not implemented yet')
    return aggregator


class CustomOutputSet:
    def __init__(self, **outputs):
        for key, value in outputs.items():
            setattr(self, key, value)


class DirectProperty(nn.Module):
    def __init__(self):
        super().__init__()

class DerivativeProperty(nn.Module):
    def __init__(self):
        super().__init__()
        self.create_graph = False  # Set by the model with train() or eval()

    def get_pairwise_force(self, outputs):
        if not hasattr(outputs, 'pairwise_force'):
            outputs.pairwise_force = -grad(
                outputs.energy, 
                outputs.disp, 
                grad_outputs=torch.ones_like(outputs.energy),
                create_graph=self.create_graph, 
                retain_graph=self.create_graph,
                )[0]
        return outputs.pairwise_force
    
class SecondDerivativeProperty(DerivativeProperty):
    def __init__(self):
        super().__init__()


class EnergyOutput(DirectProperty):
    '''
    Energy prediction

    Parameters:
        n_features (int): Number of features in the hidden layer.
        activation (nn.Module): Activation function.
    '''
    def __init__(self, n_features, activation):
        super().__init__()
        self.layers = nn.Sequential(
            nn.Linear(n_features, n_features),
            activation,
            nn.Linear(n_features, n_features),
            activation,
            nn.Linear(n_features, 1),
            )

    def forward(self, outputs):
        energy = self.layers(outputs.atom_node)
        # energy = scatter(energy, outputs.batch, dim=0, reduce='sum').reshape(-1)
        # outputs.energy = energy
        return energy

class GradientForceOutput(DerivativeProperty):
    '''
    Gradient force prediction
    '''
    def __init__(self):
        super().__init__()

    def forward(self, outputs):
        pairwise_force = self.get_pairwise_force(outputs)
        force = scatter(pairwise_force, outputs.edge_index[0], dim=0, reduce='sum', dim_size=outputs.atom_node.size(0)) - \
            scatter(pairwise_force, outputs.edge_index[1], dim=0, reduce='sum', dim_size=outputs.atom_node.size(0))
        # outputs.gradient_force = force
        return force
    
class DirectForceOutput(DirectProperty):
    '''
    Direct force prediction
    '''
    def __init__(self, n_features, activation):
        super().__init__()
        self.layers = nn.Sequential(
            nn.Linear(n_features, n_features),
            activation,
            nn.Linear(n_features, n_features),
            activation,
            nn.Linear(n_features, n_features),
            )

    def forward(self, outputs):
        force = self.layers(outputs.atom_node).unsqueeze(1) * outputs.force_node  # n_nodes, 3, n_features
        force = force.sum(dim=-1)  # n_nodes, 3
        # outputs.direct_force = force
        return force
    
class HessianOutput(SecondDerivativeProperty):
    '''
    Hessian prediction
    '''
    def __init__(self):
        super().__init__()

    def forward(self, outputs):
        hessian = torch.vmap(
            lambda vec: grad(
                -outputs.gradient_force.flatten(), 
                outputs.disp, 
                grad_outputs=vec, 
                create_graph=self.create_graph,
                retain_graph=True,
                )[0],
            )(torch.eye(outputs.gradient_force.numel()))
        hessian = hessian.reshape(*outputs.gradient_force.shape, *outputs.disp.shape)
        hessian = scatter(hessian, outputs.edge_index[0], dim=2, reduce='sum', dim_size=outputs.atom_node.size(0)) - \
            scatter(hessian, outputs.edge_index[1], dim=2, reduce='sum', dim_size=outputs.atom_node.size(0))
        # outputs.hessian = hessian
        return hessian
    
class VirialOutput(DerivativeProperty):
    '''
    Virial prediction
    '''
    def __init__(self):
        super().__init__()

    def forward(self, outputs):
        pairwise_force = self.get_pairwise_force(outputs)
        virial = outputs.disp[:, :, None] * pairwise_force[:, None, :]
        virial = scatter(virial, outputs.edge_index[0], dim=0, reduce='sum', dim_size=outputs.atom_node.size(0)) + \
            scatter(virial, outputs.edge_index[1], dim=0, reduce='sum', dim_size=outputs.atom_node.size(0))
        virial = scatter(virial, outputs.batch, dim=0, reduce='sum')
        return virial
    

class SumAggregator(nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, output, outputs):
        return scatter(output, outputs.batch, dim=0, reduce='sum').reshape(-1)

class NullAggregator(nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, output, outputs):
        return output


import torch
import torch.nn as nn
from torch_geometric.utils import scatter

def get_loss_by_string(losses):
    '''
    Get loss function by string

    Parameters:
        losses (dict): The loss function settings.
            key (str): The loss function type.
            value (dict): The loss function settings.
                weight (float): The weight for the loss function. Default: None.
                mode (str): The loss function to use. Default: 'mse'.
                    'mse': Mean squared error.
                    'mae': Mean absolute error.
                    'huber': Huber loss.
                transform (str): The transformation to apply to the data. Default: None.
                    'cos': 1 - cosine similarity.
                    'norm': Norm.

    Returns:
        main_loss (nn.Module): The main loss function for model training (back propagation) and validation (learning rate scheduling).
        eval_loss (nn.Module): The evaluation loss function for task-specific model evaluation.
    '''
    main_losses = []
    eval_losses = []
    assert losses is not None, 'losses is not defined.'
    for key, kwargs in losses.items():
        if key == 'energy':
            main_losses.append(EnergyLoss(**kwargs))
            eval_losses.append(EnergyLoss(mode='mae'))
            eval_losses.append(EnergyLoss(mode='mse'))
            eval_losses.append(EnergyPerAtomLoss(mode='mae'))
            eval_losses.append(EnergyPerAtomLoss(mode='mse'))
        elif key == 'gradient_force':
            main_losses.append(GradientForceLoss(**kwargs))
            eval_losses.append(GradientForceLoss(mode='mae'))
            eval_losses.append(GradientForceLoss(mode='mse'))
        elif key == 'direct_force':
            main_losses.append(DirectForceLoss(**kwargs))
            eval_losses.append(DirectForceLoss(mode='mae'))
            eval_losses.append(DirectForceLoss(mode='mse'))
            eval_losses.append(DirectForceLoss(mode='mae', transform='cos'))
            eval_losses.append(DirectForceLoss(mode='mse', transform='cos'))
            eval_losses.append(DirectForceLoss(mode='mae', transform='norm'))
            eval_losses.append(DirectForceLoss(mode='mse', transform='norm'))
        main_loss = lambda pred, data: sum([loss_fn(pred, data) for loss_fn in main_losses])
        eval_loss = lambda pred, data: {loss_fn.name: loss_fn(pred, data) for loss_fn in eval_losses}
    return main_loss, eval_loss


class BaseLoss(nn.Module):
    '''
    Base loss class

    Parameters:
        mode (str): The loss function to use.
            'mse': Mean squared error.
            'mae': Mean absolute error.
            'huber': Huber loss.
        weight (float): The weight for the loss function. Default: 1.
        transform (str): The transformation to apply to the data. Default: None.
            'cos': 1 - cosine similarity.
            'norm': Norm.
    '''
    def __init__(self, mode: str = 'mse', weight: float = 1, transform: str = None, **kwargs):
        super().__init__()
        self.weight = weight
        self.mode = mode
        if self.mode == 'mse':
            self.loss_fn = nn.MSELoss()
        elif self.mode == 'mae':
            self.loss_fn = nn.L1Loss()
        elif self.mode == 'huber':
            self.loss_fn = nn.HuberLoss(**kwargs)
        else:
            raise ValueError(f'loss mode {mode} not implemented')
        self.transform = transform
        if self.transform is None:
            pass
        elif self.transform == 'cos':
            self.cos = nn.CosineSimilarity(dim=-1)
            self.transform_fn = lambda x, y: (self.cos(x, y), torch.ones(x.shape[:-1], device=x.device))
        elif self.transform == 'norm':
            self.transform_fn = lambda x, y: (x.norm(dim=-1), y.norm(dim=-1))
        else:
            raise ValueError(f'transform {transform} not implemented')
        
    def forward(self, pred, data):
        if self.weight == 0:
            return 0
        value_pred, value_data = self.from_outputs(pred, data)
        if self.transform is not None:
            value_pred, value_data = self.transform_fn(value_pred, value_data)
        loss = self.loss_fn(value_pred, value_data)
        if self.weight == 1:
            return loss
        else:
            return self.weight * loss
    
    def from_outputs(self, pred, data):
        raise NotImplementedError
    

class EnergyLoss(BaseLoss):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        if self.transform is None:
            self.name = f'energy_{self.mode}'
        else:
            raise ValueError(f'transform {self.transform} for energy not implemented')

    def from_outputs(self, pred, data):
        return pred.energy, data.energy

class EnergyPerAtomLoss(BaseLoss):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        if self.transform is None:
            self.name = f'energy_per_atom_{self.mode}'
        else:
            raise ValueError(f'transform {self.transform} for energy per atom not implemented')

    def from_outputs(self, pred, data):
        n_atoms = scatter(torch.ones_like(data.z), data.batch)
        return pred.energy / n_atoms, data.energy / n_atoms

class GradientForceLoss(BaseLoss):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        if self.transform is None:
            self.name = f'gradient_force_{self.mode}'
        else:
            self.name = f'gradient_force_{self.transform}_{self.mode}'
            
    def from_outputs(self, pred, data):
        return pred.gradient_force, data.force
    
class DirectForceLoss(BaseLoss):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        if self.transform is None:
            self.name = f'direct_force_{self.mode}'
        else:
            self.name = f'direct_force_{self.transform}_{self.mode}'
            
    def from_outputs(self, pred, data):
        return pred.direct_force, data.force


import torch
from torch import nn
from torch_geometric.utils import scatter


def get_scaler_by_string(key):
    if key == 'energy':
        scaler = ScaleShift(scale=True, shift=True)
    elif key == 'gradient_force':
        scaler = ScaleShift(scale=False, shift=False)
    elif key == 'direct_force':
        scaler = ScaleShift(scale=True, shift=False)
    elif key == 'hessian':
        scaler = ScaleShift(scale=False, shift=False)
    elif key == 'virial':
        scaler = ScaleShift(scale=False, shift=False)
    else:
        raise NotImplementedError(f'Scaler type {key} is not implemented yet')
    return scaler

def set_scaler_by_string(key, scaler, stats, fit_scale=True, fit_shift=True):
    if key == 'energy':
        if fit_scale:
            scaler.set_scale(stats['energy']['scale'])
        if fit_shift:
            scaler.set_shift(stats['energy']['shift'])
    elif key == 'gradient_force':
        pass
    elif key == 'direct_force':
        if fit_scale:
            scaler.set_scale(stats['force']['scale'])
    # elif scaler.key == 'hessian':
    #     pass
    else:
        raise NotImplementedError(f'Scaler type {key} is not implemented yet')
    return scaler

class ScaleShift(nn.Module):
    '''
    Node-level scale and shift layer.
    
    Parameters:
        key (str): The key for the scaler
        scale (bool): Whether to scale the output.
        shift (bool): Whether to shift the output.
    '''
    def __init__(self, scale=True, shift=True):
        super().__init__()
        self.scale = nn.Embedding.from_pretrained(torch.ones(118 + 1, 1), freeze=False, padding_idx=0) if scale else None
        self.shift = nn.Embedding.from_pretrained(torch.zeros(118 + 1, 1), freeze=False, padding_idx=0) if shift else None

    def forward(self, output, outputs):
        '''
        Scale and shift input.

        Args:
            output (torch.Tensor): The output values.
            outputs (Data): Other output data.
        '''
        if self.scale is not None:
            output = output * self.scale(outputs.z)
        if self.shift is not None:
            output = output + self.shift(outputs.z)
        return output
    
    def set_scale(self, scale):
        self.scale.weight.data = scale.reshape(-1, 1)

    def set_shift(self, shift):
        self.shift.weight.data = shift.reshape(-1, 1)

    
    def __repr__(self):
        return f'{self.__class__.__name__}(scale={self.scale is not None}, shift={self.shift is not None})'
    


import torch
from torch_geometric.data import Data
from torch_geometric.transforms import BaseTransform
from torch_geometric.nn import radius_graph

class RadiusGraph(BaseTransform):
    r"""Creates edges based on node positions :obj:`data.pos` to all points
    within a given distance (functional name: :obj:`radius_graph`).

    Args:
        r (float): The distance.
        loop (bool, optional): If :obj:`True`, the graph will contain
            self-loops. (default: :obj:`False`)
        flow (str, optional): The flow direction when using in combination with
            message passing (:obj:`"source_to_target"` or
            :obj:`"target_to_source"`). (default: :obj:`"source_to_target"`)
        num_workers (int): Number of workers to use for computation. Has no
            effect in case :obj:`batch` is not :obj:`None`, or the input lies
            on the GPU. (default: :obj:`1`)
    """
    def __init__(
        self,
        r: float,
        loop: bool = False,
        flow: str = 'source_to_target',
        num_workers: int = 1,
    ) -> None:
        self.r = r
        self.loop = loop
        self.max_num_neighbors = 1024
        self.flow = flow
        self.num_workers = num_workers

    def forward(self, data: Data) -> Data:
        assert data.pos is not None

        if data.lattice is not None and data.lattice.max(dim=-1).values.isfinite().any():
            n_cell = (self.r / data.lattice.norm(dim=-1)).ceil().int().flatten()
            assert len(n_cell) == 3
            n_cell_tot = (2 * n_cell + 1).prod()
            shift = torch.tensor([[i, j, k] for i in range(-n_cell[0], n_cell[0] + 1) for j in range(-n_cell[1], n_cell[1] + 1) for k in range(-n_cell[2], n_cell[2] + 1)], dtype=data.pos.dtype, device=data.pos.device)
            shift = shift @ data.lattice
            shift = shift.nan_to_num()
            shifted_pos = data.pos[:, None, :] + shift  # shape: (n_node, n_cell_tot, 3)
            shifted_pos = shifted_pos.reshape(-1, 3)  # shape: (n_node * n_cell_tot, 3)
            shifted_node_index = torch.arange(data.pos.shape[0], dtype=torch.long, device=data.pos.device)[:, None].repeat(1, n_cell_tot)  # shape: (n_node, n_cell_tot)
            shifted_node_index = shifted_node_index.reshape(-1)  # shape: (n_node * n_cell_tot)
            shifted_node_isoriginal = torch.zeros(data.pos.shape[0], n_cell_tot, dtype=torch.bool, device=data.pos.device)  # shape: (n_node, n_cell_tot)
            shifted_node_isoriginal[:, n_cell_tot // 2] = True
            shifted_node_isoriginal = shifted_node_isoriginal.reshape(-1)  # shape: (n_node * n_cell_tot)
            if data.batch is not None:
                shifted_batch = data.batch[:, None].repeat(1, n_cell_tot)  # shape: (n_node, n_cell_tot)
                shifted_batch = shifted_batch.reshape(-1)  # shape: (n_node * n_cell_tot)
            else:
                shifted_batch = None
            shifted_edge_index = radius_graph(
                shifted_pos,
                self.r,
                shifted_batch,
                self.loop,
                max_num_neighbors=self.max_num_neighbors,
                flow=self.flow,
                num_workers=self.num_workers,
            )#.sort(dim=0)[0].unique(dim=1)
            shifted_edge_isoriginal = shifted_node_isoriginal[shifted_edge_index[0]]
            shifted_edge_index = shifted_edge_index[:, shifted_edge_isoriginal]
            data.edge_index = shifted_node_index[shifted_edge_index]
            data.disp = shifted_pos[shifted_edge_index[0]] - shifted_pos[shifted_edge_index[1]]
        else:
            data.edge_index = radius_graph(
                data.pos,
                self.r,
                data.batch,
                self.loop,
                max_num_neighbors=self.max_num_neighbors,
                flow=self.flow,
                num_workers=self.num_workers,
            )#.sort(dim=0)[0].unique(dim=1)
            data.disp = data.pos[data.edge_index[0]] - data.pos[data.edge_index[1]]

        return data

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}(r={self.r})'


import torch
from torch import nn
from torch_geometric.utils import scatter

from newtonnet.layers.activations import get_activation_by_string
from newtonnet.layers.scalers import get_scaler_by_string
from newtonnet.models.output import get_output_by_string, get_aggregator_by_string
from newtonnet.models.output import CustomOutputSet, DerivativeProperty


class NewtonNet(nn.Module):
    '''
    Molecular Newtonian Message Passing

    Parameters:
        n_features (int): Number of features in the latent layer. Default: 128.
        n_interactions (int): Number of message passing layers. Default: 3.
        activation (str): Activation function. Default: 'swish'.
        layer_norm (bool): Whether to use layer normalization. Default: False.
        infer_properties (list): The properties to predict. Default: [].
        representations (dict): The distance transformation functions.
    '''
    def __init__(
            self,
            n_features: int = 128,
            n_interactions: int = 3,
            activation: str = 'swish',
            layer_norm: bool = False,
            infer_properties: list = [],
            representations: nn.Module = None,
    ) -> None:

        super().__init__()
        activation = get_activation_by_string(activation)

        # embedding layer
        self.embedding_layer = EmbeddingNet(
            n_features=n_features,
            representations=representations,
            )

        # message passing
        self.interaction_layers = nn.ModuleList([
            InteractionNet(
                n_features=n_features,
                n_basis=self.embedding_layer.n_basis,
                activation=activation,
                layer_norm=layer_norm,
                ) for _ in range(n_interactions)
            ])

        # final output layer
        self.infer_properties = infer_properties
        self.output_layers = nn.ModuleList()
        self.scalers = nn.ModuleList()
        self.aggregators = nn.ModuleList()
        for key in self.infer_properties:
            output_layer = get_output_by_string(key, n_features, activation)
            self.output_layers.append(output_layer)
            if isinstance(output_layer, DerivativeProperty):
                self.embedding_layer.requires_dr = True
            scaler = get_scaler_by_string(key)
            self.scalers.append(scaler)
            aggregator = get_aggregator_by_string(key)
            self.aggregators.append(aggregator)


    # def forward(self, batch):
    def forward(self, z, disp, edge_index, batch):
        '''
        Network forward pass

        Parameters:
            batch: The input data.
                z (torch.Tensor): The atomic numbers of the atoms in the molecule. Shape: (n_nodes, ).
                disp (torch.Tensor): The displacement vectors of the atoms in the molecule. Shape: (n_edges, 3).
                edge_index (torch.Tensor): The edge index of the atoms in the molecule. Shape: (2, n_edges).
                batch (torch.Tensor): The batch of the atoms in the molecule. Shape: (n_nodes, ).

        Returns:
            outputs (dict): The outputs of the network.
        '''

        # initialize node and edge representations
        atom_node, force_node, disp_node, dir_edge, dist_edge = self.embedding_layer(z, disp)

        # compute interaction block and update atomic embeddings
        for interaction_layer in self.interaction_layers:
            atom_node, force_node, disp_node = interaction_layer(atom_node, force_node, disp_node, dir_edge, dist_edge, edge_index)

        # output net
        outputs = CustomOutputSet(z=z, disp=disp, atom_node=atom_node, force_node=force_node, edge_index=edge_index, batch=batch)
        for key, output_layer, scaler, aggregator in zip(self.infer_properties, self.output_layers, self.scalers, self.aggregators):
            output = output_layer(outputs)
            output = scaler(output, outputs)
            output = aggregator(output, outputs)
            setattr(outputs, key, output)

        return outputs
    
    def train(self, mode=True):
        '''
        Set the network to training mode
        '''
        super().train(mode)
        for output_layer in self.output_layers:
            if isinstance(output_layer, DerivativeProperty):
                output_layer.create_graph = mode


class EmbeddingNet(nn.Module):
    '''
    Embedding layer of the network

    Parameters:
        n_features (int): Number of features in the hidden layer.
        representations (dict): The distance transformation functions.
    '''
    def __init__(self, n_features, representations):

        super().__init__()

        # atomic embedding
        self.n_features = n_features
        self.node_embedding = nn.Embedding(118 + 1, n_features, padding_idx=0)

        # edge embedding
        self.norm = representations['norm']
        self.cutoff = representations['cutoff']
        self.edge_embedding = representations['radial']
        self.n_basis = self.edge_embedding.n_basis

        # requires dr
        self.requires_dr = False

    def forward(self, z, disp):

        # initialize node representations
        atom_node = self.node_embedding(z)  # n_nodes, n_features
        force_node = torch.zeros(z.size(0), 3, self.n_features, dtype=disp.dtype, device=disp.device)  # n_nodes, 3, n_features
        disp_node = torch.zeros(z.size(0), 3, self.n_features, dtype=disp.dtype, device=disp.device)  # n_nodes, 3, n_features

        # recompute distances and distance vectors
        if self.requires_dr:
            disp.requires_grad = True

        # initialize edge representations
        dist_edge, dir_edge = self.norm(disp)  # n_edges, 1; n_edges, 3
        dist_edge = self.cutoff(dist_edge) * self.edge_embedding(dist_edge)  # n_edges, n_basis

        return atom_node, force_node, disp_node, dir_edge, dist_edge
        # return atom_node, force_node, disp_node, dir_edge, cutoff_edge, rbf_edge


class InteractionNet(nn.Module):
    '''
    Message passing layer of the network

    Parameters:
        n_features (int): Number of features in the hidden layer.
        n_basis (int): Number of radial basis functions.
        activation (nn.Module): Activation function.
        layer_norm (bool): Whether to use layer normalization.
    '''
    def __init__(self, n_features, n_basis, activation, layer_norm):
        super().__init__()

        self.n_features = n_features

        # invariant message passing
        self.message_nodepart = nn.Sequential(
            nn.Linear(n_features, n_features),
            activation,
            nn.Linear(n_features, n_features),
        )
        self.message_edgepart = nn.Linear(n_basis, n_features, bias=False)
        
        self.equiv_message1 = nn.Sequential(
            nn.Linear(n_features, n_features, bias=False),
            activation,
            nn.Linear(n_features, n_features, bias=False),
        )
        self.equiv_message2 = nn.Sequential(
            nn.Linear(n_features, n_features, bias=False),
            activation,
            nn.Linear(n_features, n_features, bias=False),
        )
        
        self.equiv_update = nn.Linear(n_features, n_features, bias=False)

        # layer norm
        if layer_norm:
            self.layer_norm = nn.LayerNorm(n_features)
        else:
            self.layer_norm = None
    
    def forward(self, atom_node, force_node, disp_node, dir_edge, dist_edge, edge_index):
        # a
        message_nodepart = self.message_nodepart(atom_node)    # n_nodes, n_features
        message_edgepart = self.message_edgepart(dist_edge)    # n_edges, n_features
        message = message_edgepart * message_nodepart[edge_index[0]] * message_nodepart[edge_index[1]]    # n_edges, n_features

        inv_message1 = message    # n_nodes, n_features
        inv_update1 = scatter(inv_message1, edge_index[0], dim=0, dim_size=atom_node.size(0))    # n_nodes, n_features
        atom_node = atom_node + inv_update1    # n_nodes, n_features

        # f
        equiv_message1_invpart = self.equiv_message1(message).unsqueeze(1)    # n_edges, 1, n_features
        equiv_message1_equivpart = dir_edge.unsqueeze(2)    # n_edges, 3, 1
        equiv_message1 = equiv_message1_invpart * equiv_message1_equivpart    # n_edges, 3, n_features

        equiv_message2_invpart = self.equiv_message2(message).unsqueeze(1)    # n_edges, 1, n_features
        equiv_message2_equivpart = force_node[edge_index[1]]    # n_edges, 3, n_features
        equiv_message2 = equiv_message2_invpart * equiv_message2_equivpart    # n_edges, 3, n_features

        force_update = scatter(equiv_message1 + equiv_message2, edge_index[0], dim=0, dim_size=force_node.size(0))    # n_nodes, 3, n_features
        force_node = force_node + force_update    # n_nodes, 3, n_features

        # update energy
        inv_update2 = torch.sum(force_node * self.equiv_update(force_node), dim=1)    # n_nodes, n_features
        atom_node = atom_node + inv_update2

        # layer norm
        if self.layer_norm is not None:
            atom_node = self.layer_norm(atom_node)

        return atom_node, force_node, disp_node



import os
import pandas as pd
import torch
import shutil
import matplotlib.pyplot as plt
from tqdm import tqdm
import wandb

import torch
from torch import nn, optim
from torch.optim.lr_scheduler import ReduceLROnPlateau, LRScheduler
from torch_geometric.loader import DataLoader
from torch_geometric.nn import DataParallel


class Trainer(object):
    '''
    Trainer class for NewtonNet.

    Parameters:
        model (nn.Module): The model to train.
        loss_fns (nn.Module, nn.Module): The loss functions to use for training and evaluation. Default: None.
        optimizer (optim.Optimizer): The optimizer to use for training. Default: Adam.
        lr_scheduler (optim.lr_scheduler._LRScheduler): The learning rate scheduler to use for training. Default: ReduceLROnPlateau.
        output_base_path (str): The base path for the output directory.
        script_path (str): The path to the script that was used to start the training.
        settings_path (str): The path to the settings file that was used to start the training.
        checkpoint (dict): The checkpoint settings.
            check_log (int): The interval in epochs for logging the training progress. Default: 1.
            check_val (int): The interval in epochs for validation. Default: 1.
            check_test (int): The interval in epochs for testing. Default: 1.
            check_model (int): The interval in epochs for saving the model (must be validated first). Default: 1.
        device (torch.device): The device to use for training. Default: cpu
        train_generator (DataLoader): The training data generator. Default: None.
        val_generator (DataLoader): The validation data generator. Default: None.
        test_generator (DataLoader): The test data generator. Default: None.
        epochs (int): The number of epochs to train. Default: 100.
        clip_grad (float): The gradient clipping value. Default: 0.0.
        log_wandb (bool): Whether to use wandb for logging. Default: False.
    '''
    def __init__(
            self,
            model: nn.Module,
            loss_fns: tuple[nn.Module] = None,
            optimizer: optim.Optimizer = None,
            lr_scheduler: LRScheduler = None,
            output_base_path: str = None,
            script_path: str = None,
            settings_path: str = None,
            checkpoint: dict = {},
            device: torch.device = torch.device('cpu'),
            train_generator: DataLoader = None,
            val_generator: DataLoader = None,
            test_generator: DataLoader = None,
            epochs: int = 100,
            clip_grad: float = 0.0,
            log_wandb: bool = False,
            ):
        super().__init__()
        
        # training parameters
        self.model = model
        self.main_loss, self.eval_loss = loss_fns
        self.optimizer = optimizer
        self.lr_scheduler = lr_scheduler
        self.best_val_loss = torch.inf
        self.device = device
        self.multi_gpu = True if type(device) is list and len(device) > 1 else False
        self.model.to(self.device[0])
        if self.multi_gpu:
            self.model = DataParallel(self.model, device_ids=self.device)
        self.train_generator = train_generator
        self.val_generator = val_generator
        self.test_generator = test_generator
        self.start_epoch = 0
        self.start_step = 0
        self.epochs = epochs
        self.clip_grad = clip_grad
        self.log = pd.DataFrame()
        self.log_wandb = log_wandb

        # outputs
        if output_base_path is not None:
            self.make_subdirs(output_base_path, script_path, settings_path)
        else:
            self.output_path = None
            # self.graph_path = None
            self.model_path = None

        # checkpoints
        self.check_log = checkpoint.get('check_log', 1)
        self.check_val = checkpoint.get('check_val', 1)
        self.check_test = checkpoint.get('check_test', 1)
        self.print_layers()

    def make_subdirs(self, output_base_path, script_path, settings_path):
        assert output_base_path is not None, 'output_base_path must be specified'
        assert script_path is not None, 'script_path must be specified'
        assert settings_path is not None, 'settings_path must be specified'

        # create output directory
        path_iter = 1
        output_path = os.path.join(output_base_path, f'training_{path_iter}')
        while os.path.exists(output_path):
            path_iter += 1
            output_path = os.path.join(output_base_path, f'training_{path_iter}')
        os.makedirs(output_path)
        self.output_path = output_path
        print(f'Output directory: {output_path}')

        # create subdirectory for run scripts
        script_out = os.path.join(self.output_path, 'run_scripts')
        os.makedirs(script_out)
        shutil.copyfile(script_path, os.path.join(script_out, os.path.basename(script_path)))
        shutil.copyfile(settings_path, os.path.join(script_out, os.path.basename(settings_path)))

        # create subdirectory for computation graph
        # self.graph_path = os.path.join(self.output_path, 'graph')
        # os.makedirs(self.graph_path)

        # create subdirectory for models
        self.model_path = os.path.join(self.output_path, 'models')
        os.makedirs(self.model_path)

    def resume(self, checkpoint):
        shutil.copyfile(
            os.path.join(checkpoint, 'models', 'train_state.pt'),
            os.path.join(self.output_path, 'models', 'train_state.pt')
        )
        shutil.copyfile(
            os.path.join(checkpoint, 'models', 'best_model.pt'),
            os.path.join(self.output_path, 'models', 'best_model.pt')
        )
        shutil.copyfile(
            os.path.join(checkpoint, 'log.csv'),
            os.path.join(self.output_path, 'log.csv')
        )
        train_state = torch.load(os.path.join(self.output_path, 'models', 'train_state.pt'))
        self.start_epoch = train_state['epoch'] + 1
        self.start_step = train_state['step']
        self.model.load_state_dict(train_state['model_state_dict'])
        self.optimizer.load_state_dict(train_state['optimizer_state_dict'])
        self.lr_scheduler.load_state_dict(train_state['scheduler_state_dict'])
        self.best_val_loss = train_state['best_val_loss']
        torch.set_rng_state(train_state['rng_state'])
        self.log = pd.read_csv(os.path.join(self.output_path, 'log.csv'))

    def print_layers(self):
        print('Model:')
        print(self.model)
        print(f'total trainable parameters: {sum(p.numel() for p in self.model.parameters() if p.requires_grad)}')
        print()

    def plot_grad_flow(self, epoch):
        grads = []
        layers = []
        for name, parameter in self.model.named_parameters():
            if not parameter.requires_grad:
                continue
            if parameter.grad is not None:
                layers.append(name)
                grads.append(parameter.grad.detach().abs().mean().cpu())
            else:
                # raise ValueError(f'parameter {name} has no gradient')
                pass

        plt.figure(figsize=(16, 3))
        plt.plot(grads, '-', color='tab:blue')
        plt.xticks(range(len(layers)), layers, rotation='vertical')
        plt.xlim(-1, len(layers))
        plt.yscale('log')
        plt.xlabel('Layers')
        plt.ylabel('Average gradients')
        plt.title(f'Gradient flow - Epoch {epoch}')
        plt.grid(True)
        plt.savefig(os.path.join(self.graph_path, f'grad_flow_{epoch}.png'), dpi=300, bbox_inches='tight')
        plt.close()

    def local_log(self, log):
        log = pd.DataFrame(log, index=[0])
        self.log = pd.concat([self.log, log], ignore_index=True)
        self.log.to_csv(os.path.join(self.output_path, 'log.csv'), index=False)

    def train(self):       
        step = self.start_step
        for epoch in tqdm(range(self.start_epoch, self.epochs)):
            log_one_epoch = {'epoch': epoch, 'lr': self.optimizer.param_groups[0]['lr']}

            # training
            self.model.train()
            train_log = self.run_one_epoch(self.train_generator, step=True)
            step += len(self.train_generator)
            log_one_epoch['step'] = step
            log_one_epoch = log_one_epoch | {f'train_{key}': value for key, value in train_log.items()}

            # validation
            if epoch % self.check_val == 0 and self.val_generator is not None:
                self.model.eval()
                val_log = self.run_one_epoch(self.val_generator, step=False)
                log_one_epoch = log_one_epoch | {f'val_{key}': value for key, value in val_log.items()}

            # save test predictions
            if epoch % self.check_test == 0 and self.test_generator is not None:
                self.model.eval()
                test_log = self.run_one_epoch(self.test_generator, step=False)
                log_one_epoch = log_one_epoch | {f'test_{key}': value for key, value in test_log.items()}

            # best model
            if epoch % self.check_log == 0 and self.model_path is not None:
                if log_one_epoch['val_loss'] < self.best_val_loss:
                    self.best_val_loss = log_one_epoch['val_loss']
                    if self.multi_gpu:
                        save_model = self.model.module
                    else:
                        save_model = self.model
                    torch.save(save_model, os.path.join(self.model_path, 'best_model.pt'))
                    log_one_epoch['save_model'] = True

            # plots
            # if epoch % self.check_log == 0 and self.graph_path is not None:
            #     self.plot_grad_flow(epoch)
            if self.output_path is not None:
                self.local_log(log_one_epoch)
            if self.log_wandb:
                wandb.log(log_one_epoch)

            # learning rate decay
            if isinstance(self.lr_scheduler, ReduceLROnPlateau):
                if 'val_loss' in log_one_epoch:
                    self.lr_scheduler.step(log_one_epoch['val_loss'])
            elif isinstance(self.lr_scheduler, LRScheduler):
                self.lr_scheduler.step()

            # # loss force weight decay
            # self.main_loss.force_loss_decay()

            # checkpoint
            if epoch % self.check_log == 0 and self.model_path is not None:
                torch.save({
                        'epoch': epoch,
                        'step': step,
                        'model_state_dict': self.model.state_dict(),
                        'optimizer_state_dict': self.optimizer.state_dict(),
                        'scheduler_state_dict': self.lr_scheduler.state_dict() if self.lr_scheduler is not None else None,
                        'best_val_loss': self.best_val_loss,
                        'rng_state': torch.get_rng_state(),
                    }, os.path.join(self.model_path, 'train_state.pt'))
                
                if isinstance(self.lr_scheduler, ReduceLROnPlateau):
                    if self.optimizer.param_groups[0]['lr'] <= self.lr_scheduler.min_lrs[0]:
                        break

        print('Training finished')
        
        # load best model
        self.model = torch.load(os.path.join(self.model_path, 'best_model.pt'), weights_only=False)
        self.model.eval()
        log_one_epoch = {'epoch': 'final'}
        train_log = self.run_one_epoch(self.train_generator, step=False)
        log_one_epoch = log_one_epoch | {f'train_{key}': value for key, value in train_log.items()}
        if self.val_generator is not None:
            val_log = self.run_one_epoch(self.val_generator, step=False)
            log_one_epoch = log_one_epoch | {f'val_{key}': value for key, value in val_log.items()}
        if self.test_generator is not None:
            test_log = self.run_one_epoch(self.test_generator, step=False)
            log_one_epoch = log_one_epoch | {f'test_{key}': value for key, value in test_log.items()}
        if self.output_path is not None:
            self.local_log(log_one_epoch)
        if self.log_wandb:
            wandb.log(log_one_epoch)


    def run_one_epoch(self, generator, step=False):
        log_one_epoch = {}
        for batch in generator:
            if step:
                self.optimizer.zero_grad()
            # preds = self.model(batch)
            preds = self.model(batch.z, batch.disp, batch.edge_index, batch.batch)
            main_loss = self.main_loss(preds, batch)
            eval_loss = self.eval_loss(preds, batch)
            if step:
                main_loss.backward()
                if self.clip_grad > 0:
                    torch.nn.utils.clip_grad_norm_(self.model.parameters(), self.clip_grad)
                self.optimizer.step()
            log_one_epoch['loss'] = log_one_epoch.get('loss', 0.0) + main_loss.detach().item()
            for key, value in eval_loss.items():
                log_one_epoch[key] = log_one_epoch.get(key, 0.0) + value.detach().item()
        log_one_epoch = {key: value / len(generator) for key, value in log_one_epoch.items()}
        return log_one_epoch


import os
import os.path as osp
from tqdm import tqdm
from typing import Callable, List, Optional, Union
import numpy as np
import ase
from ase import units
from ase.io import read
units.__setattr__('kcal/mol', units.kcal / units.mol)
units.__setattr__('kJ/mol', units.kJ / units.mol)

import torch
import torch.nn as nn
from torch_geometric.data import Dataset, InMemoryDataset, Data
from torch_geometric.utils import scatter


class MolecularDataset(Dataset):
    '''
    This class is a dataset for molecular data.
    
    Args:
        root (str): The root directory where the dataset should be saved.
        transform (callable, optional): A function/transform that takes in a data object and returns a transformed version. The data object will be transformed before every access. Default: None.
        pre_transform (callable, optional): A function/transform that takes in a data object and returns a transformed version. The data object will be transformed before being saved to disk. Default: None.
        pre_filter (callable, optional): A function that takes in a data object and returns a boolean value, indicating whether the data object should be included in the final dataset. Default: None.
        force_reload (bool): Whether to re-process the dataset. Default: False.
        precision (torch.dtype): The precision of the data. Default: torch.float.
    '''
    def __init__(
        self,
        precision: torch.dtype = torch.float,
        data_length_unit: str = 'Ang',
        data_energy_unit: str = 'eV',
        **kwargs,
    ) -> None:
        self.precision = precision
        self.units = {'length': getattr(units, data_length_unit), 'energy': getattr(units, data_energy_unit)}
        super().__init__(**kwargs)

    @property
    def raw_dir(self) -> str:
        return osp.join(self.root, 'raw')

    @property
    def processed_dir(self) -> str:
        return osp.join(self.root, 'processed')

    @property
    def raw_file_names(self) -> Union[str, List[str]]:
        names = [name for name in os.listdir(self.raw_dir) if name.endswith(('.npz', '.xyz', '.extxyz'))]
        return names

    # @property
    def processed_file_names(self) -> List[str]:
        if not osp.exists(self.processed_dir):
            return []
        return [name for name in os.listdir(self.processed_dir) if name.startswith('data_') and name.endswith('.pt')]

    def process(self) -> None:
        idx = 0
        for raw_path in tqdm(self.raw_paths):
            if raw_path.endswith('.npz'):
                data_list = parse_npz(raw_path, self.pre_transform, self.pre_filter, self.precision, self.units)
            elif raw_path.endswith('.xyz') or raw_path.endswith('.extxyz'):
                data_list = parse_xyz(raw_path, self.pre_transform, self.pre_filter, self.precision, self.units)
            
            for data in data_list:
                torch.save(data, osp.join(self.processed_dir, f'data_{idx}.pt'))
                idx += 1

    def len(self) -> int:
        return len(self.processed_file_names())
    
    def get(self, idx: int) -> Data:
        return torch.load(osp.join(self.processed_dir, f'data_{idx}.pt'))
    
class MolecularInMemoryDataset(InMemoryDataset):
    '''
    This class is an in-memeory dataset for molecular data.
    
    Args:
        root (str): The root directory where the dataset should be saved.
        transform (callable, optional): A function/transform that takes in a data object and returns a transformed version. The data object will be transformed before every access. Default: None.
        pre_transform (callable, optional): A function/transform that takes in a data object and returns a transformed version. The data object will be transformed before being saved to disk. Default: None.
        pre_filter (callable, optional): A function that takes in a data object and returns a boolean value, indicating whether the data object should be included in the final dataset. Default: None.
        force_reload (bool): Whether to re-process the dataset. Default: False.
        precision (torch.dtype): The precision of the data. Default: torch.float.
    '''
    def __init__(
        self,
        precision: torch.dtype = torch.float,
        data_length_unit: str = 'Ang',
        data_energy_unit: str = 'eV',
        **kwargs,
    ) -> None:
        self.precision = precision
        self.units = {'length': getattr(units, data_length_unit), 'energy': getattr(units, data_energy_unit)}
        super().__init__(**kwargs)

        self.load(self.processed_paths[0])

    @property
    def raw_dir(self) -> str:
        return osp.join(self.root, 'raw')

    @property
    def processed_dir(self) -> str:
        return osp.join(self.root, 'processed')

    @property
    def raw_file_names(self) -> Union[str, List[str]]:
        names = [name for name in os.listdir(self.raw_dir) if name.endswith(('.npz', '.xyz', '.extxyz'))]
        return names

    # @property
    def processed_file_names(self) -> List[str]:
        return ['data.pt']

    def process(self) -> None:
        data_list = []
        data_path = self.processed_paths[0]
        for raw_path in tqdm(self.raw_paths):
            if raw_path.endswith('.npz'):
                data_list.extend(parse_npz(raw_path, self.pre_transform, self.pre_filter, self.precision, self.units))
            elif raw_path.endswith('.xyz') or raw_path.endswith('.extxyz'):
                data_list.extend(parse_xyz(raw_path, self.pre_transform, self.pre_filter, self.precision, self.units))
            
        self.save(data_list, data_path)

def parse_npz(raw_path: str, pre_transform: Callable, pre_filter: Callable, precision: torch.dtype, units: dict) -> List[Data]:
    data_list = []
    raw_data = np.load(raw_path)

    z = torch.from_numpy(raw_data['Z']).int()
    pos = torch.from_numpy(raw_data['R']).to(precision)
    lattice = torch.from_numpy(raw_data['L']).to(precision) if 'L' in raw_data else torch.ones(3, dtype=precision) * torch.inf
    if lattice.numel() == 3:
        lattice = lattice.diag()
    elif lattice.numel() == 9:
        lattice = lattice.reshape(3, 3)
    else:
        raise ValueError('The lattice must be a single 3x3 matrix for each npz file.')
    energy = torch.from_numpy(raw_data['E']).to(precision) if 'E' in raw_data else None
    force = torch.from_numpy(raw_data['F']).to(precision) if 'F' in raw_data else None

    for i in range(pos.size(0)):
        data = Data()
        data.z = z.reshape(-1) if z.dim() < 2 else z[i].reshape(-1)
        data.pos = pos[i].reshape(-1, 3) * units['length']
        data.lattice = lattice.reshape(1, 3, 3) * units['length']
        if energy is not None:
            data.energy = energy[i].reshape(1) * units['energy']
        if force is not None:
            data.force = force[i].reshape(-1, 3) * units['energy'] / units['length']

        if pre_filter is not None and not pre_filter(data):
            continue
        if pre_transform is not None:
            data = pre_transform(data)
        data_list.append(data)

    return data_list

def parse_xyz(raw_path: str, pre_transform: Callable, pre_filter: Callable, precision: torch.dtype, units: dict) -> List[Data]:
    data_list = []
    atoms_list = ase.io.read(raw_path, index=':')

    for atoms in atoms_list:
        atoms.set_constraint()
        z = torch.from_numpy(atoms.get_atomic_numbers()).int()
        pos = torch.from_numpy(atoms.get_positions(wrap=True)).to(precision)
        lattice = torch.from_numpy(atoms.get_cell().array).to(precision)
        lattice[lattice.norm(dim=-1) < 1e-3] = torch.inf
        lattice[~atoms.get_pbc()] = torch.inf
        energy = torch.tensor(atoms.get_potential_energy(), dtype=precision)
        forces = torch.from_numpy(atoms.get_forces()).to(precision)

        data = Data()
        data.z = z.reshape(-1)
        data.pos = pos.reshape(-1, 3) * units['length']
        data.lattice = lattice.reshape(1, 3, 3) * units['length']
        data.energy = energy.reshape(1) * units['energy']
        data.force = forces.reshape(-1, 3) * units['energy'] / units['length']

        if pre_filter is not None and not pre_filter(data):
            continue
        if pre_transform is not None:
            data = pre_transform(data)
        data_list.append(data)

    return data_list


class MolecularStatistics(nn.Module):
    def __init__(self) -> None:
        super().__init__()

    def forward(self, data):
        stats = {}

        z = data.z.long().cpu()
        z_unique = z.unique()
        # stats['z'] = z_unique

        batch = data.batch.cpu()

        try:
            energy = data.energy.cpu()
            formula = scatter(nn.functional.one_hot(z), batch, dim=0).to(energy.dtype)
            solution = torch.linalg.lstsq(formula, energy, driver='gelsd').solution
            energy_shifts = torch.zeros(118 + 1, dtype=energy.dtype, device=energy.device)
            energy_shifts[z_unique] = solution[z_unique]
            stds = ((energy - torch.matmul(formula, solution)).square().sum() / (formula).sum()).sqrt()
            energy_scale = torch.ones(118 + 1, dtype=energy.dtype, device=energy.device)
            energy_scale[z_unique] = stds
            stats['energy'] = {'shift': energy_shifts, 'scale': energy_scale}
        except AttributeError:
            pass
        try:
            force = data.force.norm(dim=-1).cpu()
            means = scatter(force, z, reduce='mean')
            force_scale = torch.ones(118 + 1, dtype=force.dtype, device=force.device)
            force_scale[z_unique] = means[z_unique]
            stats['force'] = {'scale': force_scale}
        except AttributeError:
            pass
        return stats



import torch
from torch import nn


def get_representation_by_string(cutoff, cutoff_network='poly', radial_network='bessel', n_basis=20):
    representations = {}
    representations['norm'] = ScaledNorm(r=cutoff)

    if cutoff_network == 'poly':
        representations['cutoff'] = PolynomialCutoff(p=9)
    elif cutoff_network == 'cos':
        representations['cutoff'] = CosineCutoff()
    else:
        raise NotImplementedError(f'The cutoff function {cutoff_network} is unknown.')
    
    if radial_network == 'bessel':
        representations['radial'] = RadialBesselLayer(n_basis=n_basis)
    else:
        raise NotImplementedError(f'The radial function {radial_network} is unknown.')
    
    return representations


class ScaledNorm(nn.Module):
    '''
    Compute scaled norm of interatomic distances.
    Based on Johannes Klicpera, Janek Grob, Stephan Gunnemann. Directional Message Passing for Molecular Graphs. ICLR 2020.

    Parameters:
        r (float): cutoff radius.
    '''
    def __init__(self, r, **kwargs):
        super().__init__()
        self.r = r

    def forward(self, disp):
        """Compute scaled norm.

        Args:
            disp (torch.Tensor): values of interatomic distance vectors.

        Returns:
            torch.Tensor: values of scaled norm.

        """
        # Compute values of scaled norm
        dist = torch.norm(disp, dim=-1, keepdim=True)
        dir = disp / dist
        dist = dist / self.r

        return dist, dir

class PolynomialCutoff(nn.Module):
    '''
    Compute polynomial cutoff function.
    Based on Johannes Klicpera, Janek Grob, Stephan Gunnemann. Directional Message Passing for Molecular Graphs. ICLR 2020.

    Parameters:
        p (int): degree of polynomial. Default: 9.

    Notes:
        y = 1 - 0.5 * (p + 1) * (p + 2) * x^p + p * (p + 2) * x^(p + 1) - 0.5 * p * (p + 1) * x^(p + 2)
        y(0) = 1
        y(1) = 0
    '''
    def __init__(self, p, **kwargs):
        super().__init__()
        self.p = p

    def forward(self, dist):
        """Compute cutoff.

        Args:
            dist (torch.Tensor): values of scaled interatomic distances.

        Returns:
            torch.Tensor: values of cutoff function.

        """
        # Compute values of cutoff function
        cutoffs = 1 \
            - 0.5 * (self.p + 1) * (self.p + 2) * dist.pow(self.p) \
            + self.p * (self.p + 2) * dist.pow(self.p + 1) \
            - 0.5 * self.p * (self.p + 1) * dist.pow(self.p + 2)

        return cutoffs


class CosineCutoff(nn.Module):
    '''
    Compute Behler cosine cutoff function.
    Copied from: https://github.com/atomistic-machine-learning/schnetpack under the MIT License.

    Notes:
        y = 0.5 * (1 + cos(pi * x))
        y(0) = 1
        y(1) = 0
    '''
    def __init__(self, **kwargs):
        super().__init__()

    def forward(self, dist):
        """Compute cutoff.

        Args:
            dist (torch.Tensor): values of scaled interatomic distances.

        Returns:
            torch.Tensor: values of cutoff function.

        """
        # Compute values of cutoff function
        cutoffs = 0.5 * (torch.cos(dist * np.pi) + 1.0)

        return cutoffs


class RadialBesselLayer(nn.Module):
    '''
    Radial Bessel functions based on the work by DimeNet: https://github.com/klicperajo/dimenet

    Parameters:
        n_basis (int): Total number of radial functions. Default: 16.

    Notes:
        y = sin(pi * r) / (pi * r)
    '''

    def __init__(self, n_basis):
        super().__init__()
        self.n_basis = n_basis
        self.frequencies = nn.Parameter(torch.arange(1, self.n_basis + 1) * torch.pi, requires_grad=False)
        self.epsilon = 1.0e-8

    def forward(self, dist):
        '''
        Compute smeared-gaussian distance values.

        Arguments:
            dist (torch.Tensor): interatomic distance values of shape (batch_size, n_atoms, n_atoms).

        Returns:
            torch.Tensor: edge embedding of shape (batch_size, n_atoms, n_atoms, n_radials).
        '''
        out = torch.sin(self.frequencies * dist) / dist #/ self.frequencies

        return out



from tqdm import tqdm

import torch
from torch.utils.data import random_split
from torch_geometric.loader import DataLoader

from newtonnet.data import MolecularStatistics


def parse_train_test(
        in_memory: bool = True,
        train_root: str = None,
        val_root: str = None,
        test_root: str = None,
        train_size: int = None,
        val_size: int = None,
        test_size: int = None,
        stats_size: int = None,
        train_batch_size: int = 32,
        val_batch_size: int = 32,
        test_batch_size: int = 32,
        **dataset_kwargs,
        ):
    '''
    Parse the training, validation, and test data.

    Args:
        in_memory (bool): Whether to load the data in memory. Default: True.
        train_path (str): The path to the training data.
        val_path (str): The path to the validation data. If None, split from the unused training data. Default: None.
        test_path (str): The path to the test data. If None, split from the unused validation data. Default: None.
        train_size (int): The size of the training set. If None, use all available data. Default: None.
        val_size (int): The size of the validation set. If None, use all available data. Default: None.
        test_size (int): The size of the test set. If None, use all available data. Default: None.
        train_batch_size (int): The batch size for training. Default: 32.
        val_batch_size (int): The batch size for validation. Default: 32.
        test_batch_size (int): The batch size for testing. Default: 32.
        stats_batch_size (int): The batch size for statistics calculation. Default: 32.
        dataset_kwargs (dict): The keyword arguments for MolecularDataset.

    Returns:
        train_gen (torch.utils.data.DataLoader): The training data loader.
        val_gen (torch.utils.data.DataLoader): The validation data loader.
        test_gen (torch.utils.data.DataLoader): The test data loader.
    '''
    # define dataset type
    if in_memory:
        from newtonnet.data import MolecularInMemoryDataset as MolecularDataset
    else:
        from newtonnet.data import MolecularDataset

    # load data
    print('Data:')
    if train_root is not None:
        train_data = MolecularDataset(root=train_root, **dataset_kwargs)
        print(f'load {len(train_data)} data from {train_root}')
    else:
        raise ValueError('train_root must be provided')
    train_size = len(train_data) if train_size is None else train_size
    train_data, left_data = random_split(train_data, [train_size, len(train_data) - train_size])
    if val_root is not None:
        val_data = MolecularDataset(root=val_root, **dataset_kwargs)
        print(f'load {len(val_data)} data from {val_root}')
    else:
        val_data = left_data
    val_size = len(val_data) if val_size is None else val_size
    val_data, left_data = random_split(val_data, [val_size, len(val_data) - val_size])
    if test_root is not None:
        test_data = MolecularDataset(root=test_root, **dataset_kwargs)
        print(f'load {len(test_data)} data from {test_root}')
    else:
        test_data = left_data
    test_size = len(test_data) if test_size is None else test_size
    test_data, left_data = random_split(test_data, [test_size, len(test_data) - test_size])
    print(f'data size (train, val, test): {len(train_data)}, {len(val_data)}, {len(test_data)}')

    # create data loader
    train_gen = DataLoader(dataset=train_data, batch_size=train_batch_size, shuffle=True)
    val_gen = DataLoader(dataset=val_data, batch_size=val_batch_size, shuffle=(len(val_data) > 0))
    test_gen = DataLoader(dataset=test_data, batch_size=test_batch_size, shuffle=(len(test_data) > 0))
    stats_gen = DataLoader(dataset=train_data, batch_size=stats_size if stats_size is not None else len(train_data), shuffle=True)
    print(f'batch size (train, val, test): {train_batch_size}, {val_batch_size}, {test_batch_size}')

    # extract data stats
    stats_calc = MolecularStatistics()
    for batch in stats_gen:
        stats = stats_calc(batch)
        break
    # print('stats:')
    # print_stats(stats)

    return train_gen, val_gen, test_gen, stats

# def process_stats(stats_raw):
#     stats = {'z': [], 'properties': {}}
#     for stat in stats_raw:
#         stats['z'].append(stat['z'])
#         for prop, prop_dict in stat['properties'].items():  # prop: 'energy', 'force', etc
#             if prop not in stats['properties']:
#                 stats['properties'][prop] = {}
#             for key, value in prop_dict.items():  # key: 'scale', 'shift'
#                 if key not in stats['properties'][prop]:
#                     stats['properties'][prop][key] = []
#                 if value.ndim > 0:
#                     value_dense = torch.full((129, ), torch.nan, dtype=value.dtype)
#                     value_dense[stat['z']] = value
#                     value = value_dense
#                 stats['properties'][prop][key].append(value)
#     stats['z'] = torch.cat(stats['z']).unique()
#     for prop, prop_dict in stats['properties'].items():
#         for key, value in prop_dict.items():
#             stats['properties'][prop][key] = torch.stack(value).nanmean(dim=0).nan_to_num()
#     return stats

# def print_stats(stats, level=1):
#     for key, value in stats.items():
#         if isinstance(value, dict):
#             print('  ' * level + f'{key}:')
#             print_stats(value, level + 1)
#         else:
#             print('  ' * level + f'{key}: {value[value != 0]}')
#     return stats


import numpy as np
from ase import Atoms
from ase.calculators.calculator import Calculator

import torch
from torch_geometric.data import Data, Batch

from newtonnet.layers.precision import get_precision_by_string
from newtonnet.layers.scalers import get_scaler_by_string
from newtonnet.models.output import get_output_by_string, get_aggregator_by_string
from newtonnet.models.output import DerivativeProperty, SecondDerivativeProperty
from newtonnet.data import RadiusGraph


##-------------------------------------
##     ML model ASE interface
##--------------------------------------
class MLAseCalculator(Calculator):
    implemented_properties = ['energy', 'free_energy', 'forces', 'hessian', 'stress']
    # note that the free_energy is not the Gibbs/Helmholtz free energy, but the potential energy in the ASE calculator, how confusing

    ### Constructor ###
    def __init__(
            self, 
            model_path: str,
            properties: list = ['energy', 'free_energy', 'forces'], 
            disagreement: str = None,
            device: str | list[str] = 'cpu',
            precision: str = 'float32',
            script: bool = False,
            trace_n_atoms: int = 0,
            **kwargs,
            ):
        """
        Constructor for MLAseCalculator for NewtonNet models

        Parameters:
            model_path (str): The path to the model.
            settings_path (str): The path to the settings.
            properties (list): The properties to be predicted. Default: ['energy', 'free_energy', 'forces'].
            disagreement (str): The type of disagreement to be calculated. Default: None.
            device (str): The device for the calculator. Default: 'cpu'.
            precision (str): The precision of the calculator. Default: 'float32'.
            script (bool): Whether to script the model. Default: False.
            trace_n_atoms (int): The number of atoms to be traced. Default: 0.
        """
        Calculator.__init__(self, **kwargs)

        if type(device) is list:
            self.device = [torch.device(item) for item in device]
        else:
            self.device = [torch.device(device)]
        self.dtype = get_precision_by_string(precision)

        self.models = []
        self.properties = properties
        if isinstance(model_path, list):
            self.models = [self.load_model(model) for model in model_path]
        else:
            self.models = [self.load_model(model_path)]
        
        self.radius_graph = RadiusGraph(self.models[0].embedding_layer.norm.r)
        
        self.script = script
        self.trace_n_atoms = trace_n_atoms
        self.disagreement = disagreement
        

    def calculate(self, atoms=None, properties=['energy', 'forces', 'stress'], system_changes=None):
        super().calculate(atoms, self.properties, system_changes)
        if isinstance(atoms, Atoms):
            atoms = [atoms]
        data = self.format_data(atoms)
        n_models, n_frames, n_atoms = len(self.models), len(atoms), len(atoms[0])

        preds = {}
        if 'energy' in self.properties:
            preds['energy'] = np.zeros((n_models, n_frames))
        if 'free_energy' in self.properties:
            preds['free_energy'] = np.zeros((n_models, n_frames))
        if 'forces' in self.properties:
            preds['forces'] = np.zeros((n_models, n_frames, n_atoms, 3))
        if 'hessian' in self.properties:
            preds['hessian'] = np.zeros((n_models, n_frames, n_atoms, 3, n_atoms, 3))
        if 'stress' in self.properties:
            preds['stress'] = np.zeros((n_models, n_frames, 6))
        for model_, model in enumerate(self.models):
            pred = model(data.z, data.disp, data.edge_index, data.batch)
            if 'energy' in self.properties:
                energy = pred.energy.cpu().detach().numpy()
                preds['energy'][model_] = energy
            if 'free_energy' in self.properties:
                energy = pred.energy.cpu().detach().numpy()
                preds['free_energy'][model_] = energy
            if 'forces' in self.properties:
                force = pred.gradient_force.cpu().detach().numpy()
                preds['forces'][model_] = force.reshape(n_frames, n_atoms, 3)
            if 'hessian' in self.properties:
                hessian = pred.hessian.cpu().detach().numpy()
                preds['hessian'][model_] = hessian.reshape(n_frames, n_atoms, 3, n_atoms, 3)
            if 'stress' in self.properties:
                virial = pred.virial.cpu().detach().numpy()
                volume = np.array([atoms[frame].get_volume() for frame in range(n_frames)])
                stress = -virial[:, [0, 1, 2, 1, 0, 0], [0, 1, 2, 2, 2, 1]] / volume[:, None] / 2
                preds['stress'][model_] = stress
            del pred

        # self.results['outlier'] = self.q_test(preds['energy'])
        for key in self.properties:
            # self.results[key] = self.remove_outlier(preds[key], self.results['outlier']).mean(axis=0)
            self.results[key] = preds[key].mean(axis=0).squeeze()

            # if self.disagreement == 'std':
            #     self.results[key + '_disagreement'] = preds[key].std(axis=0).max()
            # elif self.disagreement == 'std_outlierremoval':
            #     self.results[key + '_disagreement'] = self.remove_outlier(preds[key], self.results['outlier']).std(axis=0).max()
            # elif self.disagreement == 'range':
            #     self.results[key + '_disagreement'] = (preds[key].max(axis=0) - preds[key].min(axis=0)).max()
            # elif self.disagreement == 'values':
            #     self.results[key + '_disagreement'] = preds[key]
            del preds[key]

    def load_model(self, model):
        # TODO: Load model with only weights
        model = torch.load(model, map_location=self.device[0], weights_only=False)
        keys_to_keep = ['energy']
        for key in self.properties:
            key = {
                'energy': 'energy',
                'free_energy': 'energy',
                'forces': 'gradient_force',
                'stress': 'virial',
                'hessian': 'hessian',
            }.get(key)
            keys_to_keep.append(key)
            if key in model.infer_properties:
                continue
            model.infer_properties.append(key)
            model.output_layers.append(get_output_by_string(key))
            model.scalers.append(get_scaler_by_string(key))
            model.aggregators.append(get_aggregator_by_string(key))
        ids_to_remove = [i for i, key in enumerate(model.infer_properties) if key not in keys_to_keep]
        for i in reversed(ids_to_remove):
            model.infer_properties.pop(i)
            model.output_layers.pop(i)
            model.scalers.pop(i)
            model.aggregators.pop(i)
        model.to(self.dtype)
        model.eval()
        model.embedding_layer.requires_dr = any(isinstance(layer, DerivativeProperty) for layer in model.output_layers)
        if any(isinstance(layer, SecondDerivativeProperty) for layer in model.output_layers):
            for layer in model.output_layers:
                if isinstance(layer, DerivativeProperty):
                    layer.create_graph = True
        return model

    def format_data(self, atoms_list):
        data_list = []
        for atoms in atoms_list:
            z = torch.tensor(atoms.get_atomic_numbers(), dtype=torch.long, device=self.device[0])
            pos = torch.tensor(atoms.get_positions(wrap=True), dtype=self.dtype, device=self.device[0])
            lattice = torch.tensor(atoms.get_cell().array, dtype=self.dtype, device=self.device[0])
            lattice[~atoms.get_pbc()] = torch.inf
            data = Data(pos=pos, z=z, lattice=lattice)
            data = self.radius_graph(data)
            data_list.append(data)
        batch = Batch.from_data_list(data_list)
        return batch

    # def remove_outlier(self, data, idx):
    #     if idx is None:
    #         return data
    #     else:
    #         return np.delete(data, idx, axis=0)

    # def q_test(self, data):
    #     '''
    #     Dixon's Q test for outlier detection

    #     Parameters:
    #         data (1d array): The data to be tested.

    #     Returns:
    #         idx (int): The index to be filtered out.
    #     '''
    #     idx = np.full_like(data, np.nan)
    #     if len(data) >= 3:
    #         q_ref = { 3: 0.970,  4: 0.829,  5: 0.710, 
    #                   6: 0.625,  7: 0.568,  8: 0.526,  9: 0.493, 10: 0.466, 
    #                  11: 0.444, 12: 0.426, 13: 0.410, 14: 0.396, 15: 0.384, 
    #                  16: 0.374, 17: 0.365, 18: 0.356, 19: 0.349, 20: 0.342, 
    #                  21: 0.337, 22: 0.331, 23: 0.326, 24: 0.321, 25: 0.317, 
    #                  26: 0.312, 27: 0.308, 28: 0.305, 29: 0.301, 30: 0.290}.get(len(self.models))  # 95% confidence interval
    #         sorted_data = np.sort(data, axis=0)
    #         q_stat_min = (sorted_data[1] - sorted_data[0]) / (sorted_data[-1] - sorted_data[0])
    #         q_stat_max = (sorted_data[-1] - sorted_data[-2]) / (sorted_data[-1] - sorted_data[0])
    #         min_mask = q_stat_min > q_ref
    #         max_mask = q_stat_max > q_ref
    #         idx[min_mask] = np.argmin(data[:, min_mask], axis=0)
    #         idx[max_mask] = np.argmax(data[:, max_mask], axis=0)
    #     return idx
    


import torch
from torch import nn


def get_activation_by_string(key):
    if key == 'swish':
        activation = nn.SiLU()
    elif key == 'silu':
        activation = nn.SiLU()
    elif key == 'relu':
        activation = nn.ReLU()
    elif key == 'elu':
        activation = nn.ELU()
    elif key == 'leaky_relu':
        activation = nn.LeakyReLU()
    elif key == 'tanh':
        activation = nn.Tanh()
    elif key == 'sigmoid':
        activation = nn.Sigmoid()
    elif key == 'softplus':
        activation = nn.Softplus()
    elif key == 'gelu':
        activation = nn.GELU()
    elif key == 'ssp':
        activation = ShiftedSoftplus()
    elif key == 'swiglu':
        activation = SwiGLU()
    else:
        raise NotImplementedError("The activation function '%s' is unknown."%str(key))
    return activation


class ShiftedSoftplus(nn.Module):
    '''
    Compute shifted soft-plus activation function.
    Copied from: https://github.com/atomistic-machine-learning/schnetpack under the MIT License.

    Notes:
        y = ln(1 + e^(-x)) - ln(2)
    '''
    def __init__(self):
        super().__init__()
        self.softplus = nn.Softplus()
        self.shift = torch.log(torch.tensor(2.0))

    def forward(self, x):
        return self.softplus(x) - self.shift
    
class SwiGLU(nn.Module):
    '''
    Compute swish-gated activation function.

    Notes:
        y = gate(x) * out(x) = swish(linear(x)) * linear(x)
    '''
    def __init__(self, in_features, out_features):
        super().__init__()
        self.linear1 = nn.Linear(in_features, out_features)
        self.linear2 = nn.Linear(in_features, out_features)
        self.gate = nn.SiLU()

    def forward(self, x):
        return self.gate(self.linear1(x)) * self.linear2(x)


import os
import argparse
import yaml
import wandb

import torch
from torch_geometric.transforms import ToDevice, Compose 

from newtonnet.models import NewtonNet
from newtonnet.train import Trainer
from newtonnet.data import RadiusGraph
from newtonnet.data import parse_train_test
from newtonnet.layers.precision import get_precision_by_string
from newtonnet.layers.representations import get_representation_by_string
from newtonnet.layers.scalers import set_scaler_by_string
from newtonnet.train.loss import get_loss_by_string
from newtonnet.train.optimizer import get_optimizer_by_string, get_scheduler_by_string
# torch.autograd.set_detect_anomaly(True)

# argument parser description
parser = argparse.ArgumentParser(
    description='This is a pacakge to train NewtonNet on a given data.',
    )
parser.add_argument(
    '-c',
    '--config',
    type=str,
    help='The path to the Yaml configuration file.',
    )
parser.add_argument(
    '-r',
    '--resume',
    type=str,
    help='The path to the checkpoint to resume training.',
    )

# define arguments
args = parser.parse_args()
checkpoint = args.resume
if checkpoint is None:
    config = args.config
else:
    assert args.config is None, 'Cannot resume and train from scratch at the same time.'
    configs = [file for file in os.listdir(os.path.join(checkpoint, 'run_scripts')) if any(file.endswith(ext) for ext in ['.yaml', '.yml'])]
    assert len(configs) == 1, f'Found {len(configs)} config files in {checkpoint}.'
    config = os.path.join(checkpoint, 'run_scripts', configs[0])

# locate files
settings_path = os.path.abspath(config)
settings = yaml.safe_load(open(settings_path, 'r'))
script_path = os.path.abspath(__file__)
output_base_path = settings['general']['output']
wandb_kwargs = settings['training'].pop('wandb', None)
if wandb_kwargs is not None:
    wandb.login()
    wandb.init(**wandb_kwargs, config=settings)

# device
precision = get_precision_by_string(settings['general']['precision'])
if type(settings['general']['device']) is list:
    device = [torch.device(item) for item in settings['general']['device']]
else:
    device = [torch.device(settings['general']['device'])]

# data
torch.manual_seed(settings['general']['seed'])
transform = Compose([
    ToDevice(device[0]),
    RadiusGraph(settings['data'].pop('cutoff')),
    ])
train_gen, val_gen, test_gen, stats = parse_train_test(
    precision=precision,
    transform=transform,
    **settings['data'],
    )

# model
represenations = get_representation_by_string(
    cutoff=transform.transforms[1].r, 
    **settings['model'].pop('representation', {}),
    )
pretrained_model = settings['model'].pop('pretrained_model', None)
if pretrained_model is not None:
    model = torch.load(pretrained_model['path'], map_location=device[0])
    model.to(precision)
    if pretrained_model.get('freeze_encoder', False):
        for param in model.embedding_layer.parameters():
            param.requires_grad = False
        for param in model.interaction_layers.parameters():
            param.requires_grad = False
    if pretrained_model.get('freeze_decoder', False):
        for param in model.output_layers.parameters():
            param.requires_grad = False
    if pretrained_model.get('freeze_scaler', False):
        for param in model.scalers.parameters():
            param.requires_grad = False
else:
    model = NewtonNet(
        representations=represenations,
        **settings['model'],
        )
    model.to(device[0])
    model.to(precision)

# fit scalers
for key, scaler in zip(model.infer_properties, model.scalers):
    set_scaler_by_string(key, scaler, stats, **settings['training'].pop('fit_scalers', {}))

# loss
main_loss, eval_loss = get_loss_by_string(settings['training'].pop('loss', None))

# optimizer
trainable_params = filter(lambda p: p.requires_grad, model.parameters())
optimizer, optimizer_kwargs = settings['training'].pop('optimizer', {'adam': {}}).popitem()
optimizer = get_optimizer_by_string(optimizer, trainable_params, **optimizer_kwargs)
lr_scheduler = settings['training'].pop('lr_scheduler', None).items()
lr_scheduler = get_scheduler_by_string(lr_scheduler, optimizer)

# trainer
trainer = Trainer(
    model=model,
    loss_fns=(main_loss, eval_loss),
    optimizer=optimizer,
    lr_scheduler=lr_scheduler,
    output_base_path=output_base_path,
    script_path=script_path,
    settings_path=settings_path,
    device=device,
    train_generator=train_gen,
    val_generator=val_gen,
    test_generator=test_gen,
    log_wandb=wandb_kwargs is not None,
    **settings['training'],
    )
if checkpoint is not None:
    trainer.resume(checkpoint)
trainer.train()

print('done!')
